# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
from .....schema_classes import FieldUsageCountsClass
from .....schema_classes import UsageAggregationClass
from .....schema_classes import UsageAggregationMetricsClass
from .....schema_classes import UserUsageCountsClass


FieldUsageCounts = FieldUsageCountsClass
UsageAggregation = UsageAggregationClass
UsageAggregationMetrics = UsageAggregationMetricsClass
UserUsageCounts = UserUsageCountsClass

# fmt: on
