# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
from .....schema_classes import ExecutionRequestInputClass
from .....schema_classes import ExecutionRequestResultClass
from .....schema_classes import ExecutionRequestSignalClass
from .....schema_classes import ExecutionRequestSourceClass
from .....schema_classes import StructuredExecutionReportClass


ExecutionRequestInput = ExecutionRequestInputClass
ExecutionRequestResult = ExecutionRequestResultClass
ExecutionRequestSignal = ExecutionRequestSignalClass
ExecutionRequestSource = ExecutionRequestSourceClass
StructuredExecutionReport = StructuredExecutionReportClass

# fmt: on
